//
// <PW-Id>
//

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright 2018-2020 by Pierre Wolfers Meylan France                       //
//                                                                           //
// This library is free software. Distribution and use rights are outlined   //
// in the file "COPYING" which should have been included with this file.     //
// If this file is missing or damaged, see the license at:                   //
//                                                                           //
//    <To be define when ready>                                              //
//                                                                           //
//   This license described in this file overrides all other licenses that   //
//   might be specified in other files for this library.                     //
//                                                                           //
//   This library is free software; you can redistribute it  and/or modify   //
//   it under the terms of the GNU Lesser General Public License as publi-   //
//   shed by  the  Free Software  Foundation;  either  version 2.1  of the   //
//   License, or (at your option) any later version.                         //
//                                                                           //
//   This library  is  distributed in  the  hope that  it will  be useful,   //
//   but   WITHOUT  ANY  WARRANTY;  without even  the  implied warranty of   //
//   MERCHANTABILITY  or  FITNESS  FOR A  PARTICULAR PURPOSE.  See the GNU   //
//   Library General Public License for more details.                        //
//                                                                           //
//   You should have  received  a copy of  the  GNU Lesser General  Public   //
//   License  along with this library  (see COPYING.LIB); if not, write to   //
//   the Free Software Foundation :                                          //
//                        Inc., 675 Mass Ave, Cambridge, MA 02139, USA.      //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



//
// P.Wolfers DiaViewer Software
//

//
// Module environement to Allocate and store a large number of user Record of various sizes.
//
//


#ifndef SERVICE_R_8STK_H
#define SERVICE_R_8STK_H


#include "Service_ENV.h"


#define AREA_SIZE_B 4096
#define AREA_SIZE_W 2048
#define AREA_SIZE_L 1024



class Re8Stk {
public:
    // Define the record to store.
    typedef union {
        SInt16     sw[256];
        UInt16     uw[256];
        SInt_8     sb[512];
        UInt_8     ub[512];
        char       ch[512];
    } Record;


private:
    typedef struct AREA {
        AREA            * nxt_, // Pointer to next area.
                        * prv_; // Pointer to previous area.
        UInt16            thd_, // Position of top HeaRec.
                          top_; // Area used size.
        SInt16 tab_[AREA_SIZE_W];// Area memory.
    } Area;

    typedef struct {
        UInt_8   link_, // The Index of previous S_Record ...
                 size_; // ... and its size (in Uint32 units)
    } Header;

    typedef union {
        Header   he_;
        UInt16   lw_;   // Is here just to force alignement on 16 bits words.
    } HeadRec;

    Area                 area_; // The internal area (always allocated) is the first ...
    Area       * fst_ = &area_, // ... in the used Area and cannot be free except when ...
               * lst_ = &area_, // ... the Varobj object is destroye.
               * cur_ = &area_, // Pointer to the current Area for allocation.
               * own_ = &area_; // Pointer to last area  wher is find a Record (used by
                                // RecNb, NextUp and NextDown).

    UInt16       fre_ =      0; // Pointer to the last found element (used by RecNb, NextUp,
                                // and NextDown)

    void     URecCopy( Record *, Record *, int );

    Area   * NewArea();         // To allocate a new Area.
    void     DelArea( Area *);  // To free an Area.
    void     Init();            // Initialize the class.

public:
    void Zero();

    Re8Stk() { Init(); }
    ~Re8Stk() { Zero(); }

    Record * Push( Record *, int );
    Record * Get( int = 0 );

    int      Size( int = 0 );

    void     Remove( Record * = 0 );
    int      Pop( Record * );

    Record * RecNb( int );
    Record * NextUp();
    Record * NextDown();


};



//
// Small template class to use with various type of record (at varible sizes).
//


template <class Ty>
class Any_Re8Stk : public Re8Stk {
public:
    Any_Re8Stk <Ty> () : Re8Stk() { }
    Ty * Push( Ty * r, int sz ) { return (Ty *) Re8Stk::Push( (Record*) &r, sz ); }
    Ty * Get() { return (Ty *) Re8Stk::Get(); }
    void Remove( Ty * r = 0 ) { Re8Stk::Remove( (Record*)r ); }
    int Pop( Ty * r ) { return Re8Stk::Pop( (Record*) r ); }
    Ty * RecNb( int n ) { return (Ty *)RecNb( n ); }
};


#endif


//
// end of <PW-Id>.
//
